==> anonnet-0.0.1-20011210.tgz: anonnet/src/object.c <==
/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "common.h"
#include "log.h"
#include "object.h"
#include "crypt.h"

#ifdef __ENABLE_DEBUG
#undef object_init
void object_init(void *o,void (*destructor)(void *),void *destructor_arg,char *func) {
#define object_init(o,f,a) object_init((o),(f),(a),__func__)
#else
void object_init(void *o,void (*destructor)(void *),void *destructor_arg) {
#endif
	object_t *obj		= (object_t *)o;

	/* initialize members */
	obj->__free		= destructor;
	obj->__free_arg		= (destructor_arg == NULL)? obj : destructor_arg;


	/* initialize mutexes */
	pthread_mutexattr_init(&obj->__sync_mutexattr);
	pthread_mutexattr_init(&obj->__gc_mutexattr);
	pthread_mutexattr_init(&obj->__cond_mutexattr);

	pthread_condattr_init(&obj->__cond_attr);

	/* these are causing seg faults somehow, somewhere */
	/*pthread_mutexattr_settype(&obj->__sync_mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutexattr_settype(&obj->__gc_mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutexattr_settype(&obj->__cond_mutexattr,PTHREAD_MUTEX_DEFAULT);*/

	pthread_mutex_init(&obj->__sync_mutex,&obj->__sync_mutexattr);
	pthread_mutex_init(&obj->__gc_mutex,&obj->__gc_mutexattr);
	pthread_mutex_init(&obj->__cond_mutex,&obj->__cond_mutexattr);

	pthread_cond_init(&obj->__cond_var,&obj->__cond_attr);

	/* don't automatically retain; better to segfault */
	/* then silently leak memory */
	obj->__gc_ref_count = 0;

	/* give it a name */
	obj->object_id	= (long int)obj;

	DUMP("%s initialized object %x",func,obj->object_id);

	__objects_initialized++;

	return; /* void */

} /* end object_init */

#ifdef __ENABLE_DEBUG
#undef object_destroy
void object_destroy(void *obj,char *func) {
#define object_destroy(o) object_destroy((o),__func__)
#else
void object_destroy(void *obj) {
#endif
	//free(((object_t *)obj)->__object_name);

	DUMP("destroying object %x",((object_t *)obj)->object_id);

	/* destroy mutexes */
	pthread_mutex_destroy(&((object_t *)obj)->__sync_mutex);
	pthread_mutex_destroy(&((object_t *)obj)->__gc_mutex);
	pthread_mutex_destroy(&((object_t *)obj)->__cond_mutex);
	
	pthread_mutexattr_destroy(&((object_t *)obj)->__sync_mutexattr);
	pthread_mutexattr_destroy(&((object_t *)obj)->__gc_mutexattr);
	pthread_mutexattr_destroy(&((object_t *)obj)->__cond_mutexattr);

	__objects_destroyed++;

	return; /* void */

} /* end object_destroy */

#ifdef __ENABLE_DEBUG
#undef object_create
object_t *object_create(char *func) {
#define object_create() object_create(__func__)
#else
object_t *object_create() {
#endif
	object_t *obj;

	if ((obj = _malloc(sizeof(object_t))) == NULL)
		return NULL;

	object_init(obj,object_free,obj);

	DUMP("%s() created object %x",func,((object_t *)obj)->object_id);

	__objects_created++;
	
	return obj;

} /* end object_create */


void object_free(void *obj) {

	object_destroy(obj);

	DUMP("freeing object %x",((object_t *)obj)->object_id);

	_free(obj);

	__objects_freed++;
	
	return; /* void */

} /* end object_free */


/* NOTE: if retain is called after this object has been
   destroyed (or a release is called that will destroy it,
   then that is a programming error and not a bug here.
   a thread should retain an object before the calling
   thread that passed it continues.
   NOTE: do retain and release need their own lock? */
#ifdef __ENABLE_DEBUG
#undef object_retain
void object_retain(void *obj,char *func) {
#define object_retain(o) object_retain((o),__func__)
#else
void object_retain(void *obj) {
#endif
	DUMP("%s() retaining object %x",func,((object_t *)obj)->object_id);

	pthread_mutex_lock(&((object_t *)obj)->__gc_mutex);

	((object_t *)obj)->__gc_ref_count++;	

	pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex);
	
	return; /* void */

} /* end object_retain */
 

#ifdef __ENABLE_DEBUG
#undef object_release
void object_release(void *obj,char *func) {
#define object_release(o) object_release((o),__func__)
#else
void object_release(void *obj) {
#endif
	DUMP("%s() releasing object %x",func,((object_t *)obj)->object_id);

	if (pthread_mutex_lock(&((object_t *)obj)->__gc_mutex) != 0) {
		ERROR("could not lock object %x",((object_t *)obj)->object_id);
		abort();
	}

	((object_t *)obj)->__gc_ref_count--;

	if (pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex) != 0) {
		ERROR("could not unlock object %x",((object_t *)obj)->object_id);
		abort();
	}


	if (((object_t *)obj)->__gc_ref_count == 0) {
		if (((object_t *)obj)->__free != NULL)
			((object_t *)obj)->__free(((object_t *)obj)->__free_arg);
	}

	return; /* void */

} /* object_release */


#ifdef __ENABLE_DEBUG
#undef object_lock
void object_lock(void *obj,char *func) {
#define object_lock(o) object_lock((o),__func__)
#else
void object_lock(void *obj) {
#endif

	DUMP("%s() locking object %lx",func,((object_t *)obj)->object_id);

	if (pthread_mutex_lock(&((object_t *)obj)->__sync_mutex) != 0) {
		ERROR("could not lock object %x",((object_t *)obj)->object_id);
	}
	
	return; /* void */

} /* end object_lock */


#ifdef __ENABLE_DEBUG
#undef object_unlock
void object_unlock(void *obj,char *func) {
#define object_unlock(o) object_unlock((o),__func__)
#else
void object_unlock(void *obj) {
#endif

	DUMP("%s() unlocking object %x",func,((object_t *)obj)->object_id);

	if (pthread_mutex_unlock(&((object_t *)obj)->__sync_mutex) != 0) {
		ERROR("could not unlock object %x",((object_t *)obj)->object_id);
	}

	return; /* void */

} /* end object_unlock */


void object_wait(void *obj) {

	//pthread_mutex_lock(&((object_t *)obj)->__cond_mutex);

	pthread_cond_wait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__sync_mutex);

	//pthread_mutex_unlock(&((object_t *)obj)->__cond_mutex);

	return; /* void */

} /* end object_wait */


void object_timedwait(void *obj,const struct timespec *tp) {

	//pthread_mutex_lock(&((object_t *)obj)->__cond_mutex);

	pthread_cond_timedwait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__sync_mutex,tp);

	//pthread_mutex_unlock(&((object_t *)obj)->__cond_mutex);

	return; /* void */

} /* end object_timedwait */


void object_signal(void *obj) {

	pthread_cond_signal(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_signal */


void object_broadcast(void *obj) {

	pthread_cond_broadcast(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_broadcast */


void object_tracker() {
	DUMP("objects initialized/destroyed: %i/%i",__objects_initialized,__objects_destroyed);
	DUMP("objects created/freed: %i/%i",__objects_created,__objects_freed);
}

==> anonnet-0.0.1-20021020.tgz: anonnet/src/object.c <==
/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "common.h"
#include "log.h"
#include "object.h"
#include "crypt.h"
#include "queue.h"


/*
 * object list management
 */
pthread_mutex_t _objects_mutex	= PTHREAD_MUTEX_INITIALIZER;
LIST_HEAD(,object_t) _objects	= LIST_HEAD_INITIALIZER(struct _objects);



#ifdef __ENABLE_DEBUG
#undef object_init
void object_init(void *o,void (*destructor)(void *),void *destructor_arg,char *func) {
#define object_init(o,f,a) object_init((o),(f),(a),__func__)
#else
void object_init(void *o,void (*destructor)(void *),void *destructor_arg) {
#endif
	object_t *obj		= (object_t *)o;

	/* initialize members */
	obj->__free		= destructor;
	obj->__free_arg		= (destructor_arg == NULL)? obj : destructor_arg;


	/* initialize mutexes */
	pthread_mutexattr_init(&obj->__sync_mutexattr);
	pthread_mutexattr_init(&obj->__gc_mutexattr);

	pthread_condattr_init(&obj->__cond_attr);

	pthread_mutex_init(&obj->__sync_mutex,&obj->__sync_mutexattr);
	pthread_mutex_init(&obj->__gc_mutex,&obj->__gc_mutexattr);

	pthread_cond_init(&obj->__cond_var,&obj->__cond_attr);

	/* don't automatically retain; better to segfault */
	/* than to silently leak memory */
	obj->__gc_ref_count = 0;

	/* register it */
	pthread_mutex_lock(&_objects_mutex);
	LIST_INSERT_HEAD(&_objects,obj,le);
	pthread_mutex_unlock(&_objects_mutex);


	DUMP("%s initialized object %p",func,(void *)obj);

	__objects_initialized++;

	return; /* void */

} /* end object_init */

#ifdef __ENABLE_DEBUG
#undef object_destroy
void object_destroy(void *obj,char *func) {
#define object_destroy(o) object_destroy((o),__func__)
#else
void object_destroy(void *obj) {
#endif
	//free(((object_t *)obj)->__object_name);

	DUMP("destroying object %p",obj);

	/* destroy mutexes */
	pthread_mutex_destroy(&((object_t *)obj)->__sync_mutex);
	pthread_mutex_destroy(&((object_t *)obj)->__gc_mutex);

	pthread_mutexattr_destroy(&((object_t *)obj)->__sync_mutexattr);
	pthread_mutexattr_destroy(&((object_t *)obj)->__gc_mutexattr);

	/* un-register it */
	pthread_mutex_lock(&_objects_mutex);
	LIST_REMOVE(((object_t *)obj),le);
	pthread_mutex_unlock(&_objects_mutex);

	__objects_destroyed++;

	return; /* void */

} /* end object_destroy */

#ifdef __ENABLE_DEBUG
#undef object_create
object_t *object_create(char *func) {
#define object_create() object_create(__func__)
#else
object_t *object_create() {
#endif
	object_t *obj;

	if ((obj = _malloc(sizeof(object_t))) == NULL)
		return NULL;

	object_init(obj,object_free,obj);

	DUMP("%s() created object %p",func,obj);

	__objects_created++;
	
	return obj;

} /* end object_create */


void object_free(void *obj) {

	object_destroy(obj);

	DUMP("freeing object %p",obj);

	_free(obj);

	__objects_freed++;
	
	return; /* void */

} /* end object_free */


/* NOTE: if retain is called after this object has been
   destroyed (or a release is called that will destroy it,
   then that is a programming error and not a bug here.
   a thread should retain an object before the calling
   thread that passed it continues.
   NOTE: do retain and release need their own lock? */
#ifdef __ENABLE_DEBUG
#undef object_retain
void object_retain(void *obj,char *func) {
#define object_retain(o) object_retain((o),__func__)
#else
void object_retain(void *obj) {
#endif
	DUMP("%s() retaining object %p",func,obj);

	pthread_mutex_lock(&((object_t *)obj)->__gc_mutex);

	((object_t *)obj)->__gc_ref_count++;	

	pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex);
	
	return; /* void */

} /* end object_retain */
 

#ifdef __ENABLE_DEBUG
#undef object_release
void object_release(void *obj,char *func) {
#define object_release(o) object_release((o),__func__)
#else
void object_release(void *obj) {
#endif
	DUMP("%s() releasing object %p",func,obj);

	if (pthread_mutex_lock(&((object_t *)obj)->__gc_mutex) != 0) {
		ERROR("could not lock object %p",obj);
		abort();
	}

	((object_t *)obj)->__gc_ref_count--;

	if (pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex) != 0) {
		ERROR("could not unlock object %p",obj);
		abort();
	}


	if (((object_t *)obj)->__gc_ref_count == 0) {
		if (((object_t *)obj)->__free != NULL)
			((object_t *)obj)->__free(((object_t *)obj)->__free_arg);
	}

	return; /* void */

} /* object_release */


#ifdef __ENABLE_DEBUG
#undef object_lock
void object_lock(void *obj,char *func) {
#define object_lock(o) object_lock((o),__func__)
#else
void object_lock(void *obj) {
#endif

	DUMP("%s() locking object %p",func,obj);

	if (pthread_mutex_lock(&((object_t *)obj)->__sync_mutex) != 0) {
		ERROR("could not lock object %p",obj);
	}
	
	return; /* void */

} /* end object_lock */


#ifdef __ENABLE_DEBUG
#undef object_unlock
void object_unlock(void *obj,char *func) {
#define object_unlock(o) object_unlock((o),__func__)
#else
void object_unlock(void *obj) {
#endif

	DUMP("%s() unlocking object %p",func,obj);

	if (pthread_mutex_unlock(&((object_t *)obj)->__sync_mutex) != 0) {
		ERROR("could not unlock object %p",obj);
	}

	return; /* void */

} /* end object_unlock */


void object_wait(void *obj) {

	pthread_cond_wait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__sync_mutex);

	return; /* void */

} /* end object_wait */


void object_timedwait(void *obj,const struct timespec *tp) {

	pthread_cond_timedwait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__sync_mutex,tp);

	return; /* void */

} /* end object_timedwait */


void object_signal(void *obj) {

	pthread_cond_signal(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_signal */


void object_broadcast(void *obj) {

	pthread_cond_broadcast(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_broadcast */


void object_tracker() {
	DUMP("objects initialized/destroyed: %i/%i",__objects_initialized,__objects_destroyed);
	DUMP("objects created/freed: %i/%i",__objects_created,__objects_freed);
}


bool is_object(void *obj) {
	object_t *o;

	pthread_mutex_lock(&_objects_mutex);

	LIST_FOREACH(o,&_objects,le) {
		if (o == ((object_t *)obj)) {
			pthread_mutex_unlock(&_objects_mutex);
			return true;
		}
	}

	pthread_mutex_unlock(&_objects_mutex);

	return false;
}
